/* r.dst.bpa */
/* Part of the GRASS Dempster-Shafer Predictive Modelling Toolkit (DST-PMT) */

/* Purpose: 	Automatically derrive BPAs (Basic Probability Assignments)
				for a predictive model of site locations using Dempster-Shafer's
				theory of evidence.
				
 				Needs two files that constitute one source of evidence: 
				(a) a fully classified input raster map (integer format)
				(b) a list of sites
				
				The difference in relative class frequencies between the entire raster map
				and the cells that lie under the sites is taken to be evidence for one of two
				mutually exclusive hypotheses:
				(a) H1 = a site is present in this cell
				(b) H2 = no site is present
				Additionally, the uncertainty hypotheses  
				{H1,H2} = H1 OR H2
				can be strengthened by independently transferring a percentage of 				
				the probability masses of each source of evidenc.
				
				The output consists of three floating point raster maps,
				encoding the BPAs for each of the three hypotheses in the
				interval [0..1].
				The output map names will be suffixed '.SITE', '.NOSITE' and
				'.SITE_NOSITE', respectively.

	This file is licensed under the GPL (version 2 or later at your option)
 	see: http://www.gnu.org/copyleft/gpl.html* /

 	(c) Benjamin Ducke 2004
	benducke@compuserve.de

*/



#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "gis.h"
#include "site.h"
#include "../../GT/rowcache.h"
#include "../../GT/cat_engine.h"

#define PROGVERSION 1.0

/* program will issue a warning, if samples are less than MIN_SAMPLE_SIZE */
#define MIN_SAMPLE_SIZE 100

typedef struct
{
	float a;
	float b;
	float a_b;
}
bpa_struct;

unsigned int WARN_ON = 0;
unsigned int WARN_N_TO_SMALL = 0;

extern int errno;
int cachesize;

char tmp_file[4096];
int debug_mode = 0;		/* 1 to enable writing debug output to logfile */
FILE *logfile;


/* This calculates the actual BPA values */
/* h1 = a%/b% = "site" (category percentage under sites) */
/* h2 = b%/a% = "no site" (category percentage in input map) */
/* h3 = "no site OR site" (uncertainty hypothesis) */
/* h3 can be determined by specifying an uncertainty percentage */
/* a = share_smp; b = share_nat */
/* int i and char *label are passed so that some information can be written
   to the logfile */
void
dst_calc (float *a, float *b, float *a_b,  
			float input_uncertainty, float site_uncertainty,
			int i, char *label)
{
	float h1, h2, h1_2, norm;
	
	h1 = *a / *b;
	if ( *b == 0 ) h1 = 0;
	h2 = *b / *a;
	if ( *a == 0 ) h2 = 0;	

	/* subtract user-defined amounts of uncertainty */
	h1 = h1 - ((h1/100) * site_uncertainty);
	h2 = h2 - ((h2/100) * input_uncertainty);
	if (site_uncertainty == 100) {
		h1 = 0;
	}
	if (input_uncertainty == 100) {
		h2 = 0;
	}
	h1_2 = ((h2/100) * input_uncertainty) + ((h1/100) * site_uncertainty);	
	if ((input_uncertainty == 100) && (site_uncertainty == 100)) {
		h1_2 = 1;
	}	
	
	/* normalise evidence to sum to 1 */	
	norm = h1 + h2 + h1_2;
	if (norm == 0) {
		h1 = 0;
		h2 = 0;
		h1_2 = 1;
	} else {
		h1 = h1/norm;
		h2 = h2/norm;
		h1_2 = h1_2/norm;
	}
			
	/* round results to 3-digits precision (good enough) */	
	h1_2 = (rintf (h1_2 * 1000)) / 1000;
	h1 = (rintf (h1 * 1000)) / 1000;
	h2 = (rintf (h2 * 1000)) / 1000;
	
	if (debug_mode)
	{
		fprintf (logfile, "%i\t%4.2f\t%4.2f\t%4.2f\t%s\n",
			 i,h1, h2, h1_2,label);
	}

	/* check, if everything still sums to 1 */
	/* checks results against an error margin of +/- 0.001 */
	/* everything else is rounding errors. */
	if ((rintf (h1_2 + h1 + h2) < 0.999)
	    || (rintf (h1_2 + h1 + h2) > 1.001))
	{
		G_warning ("Evidence does not sum to 1.: %.2f",h1_2 + h1 + h2);		
	}

	/* return results */
	*a_b = h1_2;
	*a = h1;
	*b = h2;
}


/* process input map and write BPA values to output map */
int
write_BPAs (char *coverage_answer, char* mapset, char *sites_answer,
		    char *result_answer, float input_uncertainty, float sites_uncertainty, 
			int show_progress)
{
	CELL *cellbuf;
	FCELL *fcellbuf;
	struct Categories *cats;
	struct Cell_head region;
	struct Colors colors;
	FCELL from, to; /* for color ramp */
	GT_Row_cache_t *cache;		
	int nrows, ncols;
	int row, col;
	int fd, fd_out_a, fd_out_b, fd_out_a_b;
	FILE *fp;
	FILE *fp_sites;
	int n = 1; /* number of classes */
	int i = 1;
	long total = 0; /* for cell statistics */
	long count = 0;
	float *share_nat = NULL;
	float *share_smp = NULL;
	bpa_struct *bpa; /* stores BPA values */
	/* temporary variables for DST quantification */
	float a, b, a_b;
	int error;
	char result_map[255] = "";
	char **cats_description; /* category labels */
	long *cat_count; /* category counts */
	int *site_count;
	long null_count; /* number of NULL cells */
	long nocat_count; /* number of cells that do not fall into the category range [0 .. n] */
	/* needed for analysing sites file */
	Site *sample;
	int dims;
	int cat;
	int strs;
	int dbls;
	int n_sites = 0;
	int j = 0;
	int raster_val = 0;
	long row_idx;
	long col_idx;	

	cats = G_malloc (sizeof (struct Categories));
	G_get_window (&region);
	nrows = G_window_rows ();
	ncols = G_window_cols ();	
	
	/* STEP 1: Calculate ranges and number of categories in input maps */
	
	/* open evidence raster map for reading */
	fd = G_open_cell_old (coverage_answer, G_find_cell (coverage_answer, ""));	
	if (fd < 0) {
		G_fatal_error ("Could not open coverage map for reading!\n");
	}
		
	/* get number of categories */
	G_read_cats (coverage_answer, mapset, cats);
	n = G_number_of_cats (coverage_answer, mapset) + 1;
	if (n < 2) {
		G_fatal_error ("Could not find at least two categories in input map.");
	}
	if (debug_mode)
	{
		fprintf (logfile, "Found %i categories in input map.\n", n);
		fflush (logfile);
	}
	
	/* get category labels and counts */
	cats_description = GT_get_labels (coverage_answer,mapset);
	if (cats_description == NULL) {
		G_fatal_error ("Could not read category labels from input map.");
	}	
	cat_count = GT_get_c_counts (coverage_answer,mapset, show_progress);
	if (cat_count == NULL) {
		G_fatal_error ("Could not count categories in input map.");
	}
		
	
	/* calculate percentual shares of coverage cats and store them */
	share_nat = (float *) G_malloc (n * sizeof (float));
	for (i = 0; i < n; i++) {
		share_nat[i] = 0;
		total = total + cat_count[i];
	}	
	for (i = 0; i < n; i++) {
		share_nat[i] = (float) cat_count[i] / total * 100;
	}
	if (debug_mode)
	{
		fprintf (logfile,"\nCategories in input map '%s@%s':\n", coverage_answer, mapset);	
		fprintf (logfile,"Cat.\tCount\t(%%)\tDescription\n");
		for (i = 0; i < n; i++) {
			fprintf (logfile, "%i\t%i\t%5.2f\t%s\n", i, cat_count[i],share_nat[i],
					cats_description[i]);
		}
		fflush (logfile);
	}

	/* STEP 2: count categories under site samples */
	
	/* open sites file with samples */
	fp_sites = G_sites_open_old (sites_answer, G_find_sites (sites_answer, ""));
	if (fp_sites == NULL) {
		G_fatal_error
			("Could not open site list file with samples.");
	}

	error = G_site_describe (fp_sites, &dims, &cat, &strs, &dbls);
	if (error < 0) {
		G_fatal_error
			("Could not determine format of sample site list.");
	}
	sample = G_site_new_struct (cat, dims, strs, dbls);
	if (sample == NULL) {
		G_fatal_error
			("Could not reproduce format of sample site list.");
	}
		
	/* calculate total number of site samples */
	while (G_site_get (fp_sites, sample) == 0)
	{
		n_sites++;
	}
	/* rewind sites file */
	fclose (fp_sites);
	fp_sites = G_sites_open_old (sites_answer, G_find_sites (sites_answer, ""));
	
	/* create array to store sites<->categories counts */
	site_count = G_calloc (n, sizeof(int));
	/* attach a cache to raster map */
	cache = (GT_Row_cache_t *) G_malloc (sizeof (GT_Row_cache_t));
	GT_RC_open (cache, cachesize, fd, CELL_TYPE);		
	total = 0; i = 0;
	if (show_progress) {
		fprintf (stdout, "Counting categories under sites:\n");
	}
	while (G_site_get (fp_sites, sample) == 0)
	{
		/* check if site in current region */
		if ((sample->east >= region.west) &&
		    (sample->east <= region.east) &&
		    (sample->north >= region.south) &&
		    (sample->north <= region.north))
		{				
			/* get raster row with same northing as sample and perform
			 * quantification */
			row_idx =
				(long) G_northing_to_row (sample->
							  north,
							  &region);
			col_idx =
				(long) G_easting_to_col (sample->east,
							 &region);
			cellbuf = GT_RC_get (cache, row_idx);
			/* now read the raster value under the current site */
			if (G_is_c_null_value (&cellbuf[col_idx]) == 0)
			{
				site_count [cellbuf[col_idx]]++;
				total++;				
				i ++; /* i keeps track of samples on non-null coverage inside the current region */
			}
		}
		/* update progress display */
		j ++;
		if (show_progress) {
			G_percent (j, n_sites, 1);
			fflush (stdout);
		}
	}		
	/* allocate an array to store percentages of cats under sites */
	share_smp = (float *) G_malloc (n * sizeof (float));
	/* calculate percentages */
	for (j = 0; j < n; j++) {
		share_smp[j] = (float) site_count[j] / total * 100;
	}	
	if (show_progress) {	
		fprintf (stdout, "\n");
		fflush (stdout);
	}
	/* warn if sample is too small */
	if ( i < MIN_SAMPLE_SIZE ) {
		G_warning ("Sample size is smaller than %i.\n",MIN_SAMPLE_SIZE);
		if (debug_mode) {
			WARN_ON = 1;
			WARN_N_TO_SMALL = 1;
		}
	}		
	
	if (debug_mode)
	{
		fprintf (logfile, "\nCategories under %i sites in '%s':\n", i, sites_answer);
		fprintf (logfile,"Cat.\tCount\t(%%)\tDescription\n");
		for (i=0; i < n; i ++) {
			fprintf (logfile, "%i\t%i\t%5.2f\t%s\n", i, site_count[i], share_smp[i],
						cats_description[i]);
		}
		fflush (logfile);
	}
	
	/* now do the DST quantification */
	/* a = site; b = no site; a_b = could be both */
	bpa = G_malloc (n * sizeof (bpa_struct));
	
	if (debug_mode) {
		fprintf (logfile,"\nBPA quantifications for categories:\n");
		fprintf (logfile,"(uncertainty/sites=%.2f%% uncertainty/area=%.2f%%)\n",
				sites_uncertainty, input_uncertainty);
		fprintf (logfile,"Cat.\t{H1}\t{H2}\t{H1,H2}\tDescription*\n");
	}
	
	for (i = 0; i < n; i++)
	{
		a = share_smp[i];
		b = share_nat[i];
		a_b = 0;
		/* calculate the actual BPA value */
		dst_calc (&a, &b, &a_b, input_uncertainty, sites_uncertainty, 
					i, cats_description[i]);
		bpa[i].a = a;
		bpa[i].b = b;
		bpa[i].a_b = a_b;
	}
	
	if (debug_mode) {
		fprintf (logfile,"\n*H1 = 'site present', H2 = 'no site present'\nH1,H2 = 'site present' OR 'no site present' (uncertainty)\n");
	}	
	
	/* open result maps */
	sprintf (result_map, "%s.SITE", result_answer);
	fd_out_a = G_open_raster_new (result_map, FCELL_TYPE);
	sprintf (result_map, "%s.NOSITE", result_answer);
	fd_out_b = G_open_raster_new (result_map, FCELL_TYPE);
	sprintf (result_map, "%s.SITE_NOSITE", result_answer);
	fd_out_a_b = G_open_raster_new (result_map, FCELL_TYPE);
	if ((fd_out_a < 1) || (fd_out_b < 1) || (fd_out_a_b < 1))
	{
		G_fatal_error ("Could not open result map for writing!\n");
	}
	
	if (show_progress) {
		fprintf (stdout, "Writing bpa quantifications: \n");
	}	
	fcellbuf = G_allocate_raster_buf(FCELL_TYPE);	
	for (row = 0; row < nrows; row++) {
		cellbuf = GT_RC_get (cache, row);
		/* parse one row of raster data and write bpas */
		for (col = 0; col < ncols; col++) {
			if (G_is_c_null_value (&cellbuf[col]) == 0) {
				fcellbuf[col] = bpa[cellbuf[col]].a;
			} else { /* copy null-value */				
				G_set_f_null_value (&fcellbuf[col], 1);
			}
		}		
		G_put_raster_row (fd_out_a, fcellbuf, FCELL_TYPE);
		
		for (col = 0; col < ncols; col++) {
			if (G_is_c_null_value (&cellbuf[col]) == 0) {
				fcellbuf[col] = bpa[cellbuf[col]].b;
			} else { /* copy null-value */				
				G_set_f_null_value (&fcellbuf[col], 1);
			}
		}
		G_put_raster_row (fd_out_b, fcellbuf, FCELL_TYPE);
				
		for (col = 0; col < ncols; col++) {
			if (G_is_c_null_value (&cellbuf[col]) == 0) {
				fcellbuf[col] = bpa[cellbuf[col]].a_b;
			} else { /* copy null-value */				
				G_set_f_null_value (&fcellbuf[col], 1);
			}
		}
		G_put_raster_row (fd_out_a_b, fcellbuf, FCELL_TYPE);
		
		if (show_progress) {
			G_percent (row, nrows-1, 1);
			fflush (stdout);
		}
	}
	if (show_progress) {
		fprintf (stdout, "\n");
		fflush (stdout);
	}
	
	/* we can now close the coverage map ... */
	/* close cache first */
	GT_RC_close (cache);
	/* now close the file */		
	G_close_cell (fd_out_a);
	G_close_cell (fd_out_b);
	G_close_cell (fd_out_a_b);
	
	/* write some nice color ramps */
	G_init_colors (&colors);
	from = 0;
	to = 1;
	G_add_f_raster_color_rule (&from, 100, 100, 100, &to, 255, 0, 0, &colors);
	sprintf (result_map, "%s.SITE", result_answer);
	G_write_colors (result_map, mapset, &colors);
	G_add_f_raster_color_rule (&from, 100, 100, 100, &to, 0, 255, 0, &colors);
	sprintf (result_map, "%s.NOSITE", result_answer);
	G_write_colors (result_map, mapset, &colors);
	G_add_f_raster_color_rule (&from, 100, 100, 100, &to, 255, 255, 0, &colors);
	sprintf (result_map, "%s.SITE_NOSITE", result_answer);	
	G_write_colors (result_map, mapset, &colors);
	
	/* clean up fp map stuff */
	G_free (cellbuf);
	G_free (fcellbuf);
	G_free (share_nat);
	G_free (share_smp);
	G_free (cat_count);
}


/* MAIN */
int
main (int argc, char *argv[])
{
	struct GModule *module;
	struct
	{
		struct Option *input;
		struct Option *sites;
		struct Option *input_uncertainty;
		struct Option *sites_uncertainty;
		struct Option *output;			
		struct Option *logfile;
		struct Option *cachesize;
	}
	parm;
	
	struct
	{
		struct Flag *quiet;
	}
	flag;
	
	char result_str[255] = "bpa.";
	char *mapset;
	int show_progress = 1; /* enable progress display by default */
	time_t systime;
	clock_t proctime;
	unsigned long timeused;
	unsigned int days, hours, mins, secs;
	struct History hist;	
	/* these vars are used to store information about the input map */
	int cats; /* number of categories in input map */
	char **cats_description; /* category labels */
	long *cat_count; /* category counts */
	long null_count; /* number of NULL cells */
	long nocat_count; /* number of cells that do not fall into the category range [0 .. n] */	

	/* setup some basic GIS stuff */
	G_gisinit (argv[0]);
	module = G_define_module ();
	module->description = "Produce BPA raster maps for predictive models.";
	/* do not pause after a warning message was displayed */
	G_sleep_on_error (0);

	/* Parameters: */
	/* input = a map that shows the natural (complete) distribution */
	/* of a feature. */
	parm.input = G_define_option ();
	parm.input->key = "raster";
	parm.input->type = TYPE_STRING;
	parm.input->required = YES;
	parm.input->description = "Name of GRASS raster evidence map.";

	/* site map with known sites */
	parm.sites = G_define_option ();
	parm.sites->key = "sites";
	parm.sites->type = TYPE_STRING;
	parm.sites->required = YES;
	parm.sites->description = "Name of GRASS sites evidence map.";
		
	/* percentage of uncertainty to include */
	parm.input_uncertainty = G_define_option ();
	parm.input_uncertainty->key = "raster_uncertainty";
	parm.input_uncertainty->type = TYPE_DOUBLE;
	parm.input_uncertainty->answer = "0";
	parm.input_uncertainty->required = NO;
	parm.input_uncertainty->options = "0-100";
	parm.input_uncertainty->description = 
		"Percentage of evidence mass in raster map to shift to 'uncertainty' hypothesis.";
		
	parm.sites_uncertainty = G_define_option ();
	parm.sites_uncertainty->key = "sites_uncertainty";
	parm.sites_uncertainty->type = TYPE_DOUBLE;
	parm.sites_uncertainty->answer = "0";
	parm.sites_uncertainty->required = NO;
	parm.sites_uncertainty->options = "0-100";
	parm.sites_uncertainty->description = 
		"Percentage of evidence mass in sites map to shift to 'uncertainty' hypothesis.";
		

	/* name of output map */
	parm.output = G_define_option ();
	parm.output->key = "output";
	parm.output->type = TYPE_STRING;
	parm.output->answer = "bpa.<raster>";
	parm.output->required = NO;
	parm.output->description =
		"Output raster map name. Overwrites existing!";

	/* Name of logfile. If given, Output will be logged */
	/* to this file. */
	parm.logfile = G_define_option ();
	parm.logfile->key = "logfile";
	parm.logfile->type = TYPE_STRING;
	parm.logfile->answer = "none";
	parm.logfile->required = NO;
	parm.logfile->description = "Name of file to log output.";
	
	/* number of lines to store in cache */
	parm.cachesize = G_define_option ();
	parm.cachesize->key = "cachesize";
	parm.cachesize->type = TYPE_INTEGER;
	parm.cachesize->answer = "-1";
	parm.cachesize->required = NO;
	parm.cachesize->description = "Number of raster rows to store in cache (-1 for auto)";

	/* quiet operation? */
	flag.quiet = G_define_flag ();
	flag.quiet->key = 'q';
	flag.quiet->description = "Disable on-screen progress display.";

	G_disable_interactive ();
	
	/* parse command line */
	if (G_parser (argc, argv))
	{
		exit (-1);
	}
	
	/* check for 'quiet' flag */
	if ( flag.quiet->answer ) {
		show_progress = 0;
	}	

	mapset = G_calloc (255, sizeof (char));
	/* check if input map is a fully categorised raster map */
	mapset = G_find_cell (parm.input->answer,"");
	if ( mapset == NULL) {
		G_fatal_error ("The input map does not exist in the current database.");
	}
	if (G_raster_map_is_fp (parm.input->answer, mapset)) {
		G_fatal_error ("Input map is a floating point map.\nOnly integer maps are allowed as input maps.");		
    }
	cats = GT_get_stats (parm.input->answer,mapset,&null_count, &nocat_count, show_progress);
	if ( cats < 0 ) {
		G_fatal_error ("Could not stat input map. Do you have read access?");
	}
	if ( cats == 0 ) {
		G_fatal_error ("No categories defined in input map.");		
	}
	if ( nocat_count > 0 ) {
		G_fatal_error ("Input map contains cells that do not map to categories.\nUse 'r.categorise' to produce a fully categorised input map.");
	}
	
	/* check if sample sites file exists */
	if (G_find_file ("site_lists", parm.sites->answer, "") == NULL)
	{
		G_fatal_error
			("No sample/sites list exists with the given name!");
	}

	/* construct default output filename */
	if (strcmp ("bpa.<raster>", parm.output->answer) == 0)
	{
		sprintf (result_str, "bpa.%s", parm.input->answer);
		parm.output->answer = result_str;
	}

	/* check if a legal file name was given for output map */
	/* this just overwrites existing maps! */
	if (G_legal_filename (parm.output->answer) == -1)
	{
		G_fatal_error
			("Name chosen for result map is not a valid file name!");
	}

	/* check if logfile is wanted and ... */
	if (strcmp (parm.logfile->answer, "none") != 0)
	{
		/* ... check if a legal file name was given for logfile */
		if (G_legal_filename (parm.logfile->answer) == -1)
		{
			G_fatal_error
				("Name chosen for logfile is not a valid file name!");
		}
	}		
	
	/* set cachesize */
	cachesize = atoi (parm.cachesize->answer);
	if ( (cachesize<-1) || (cachesize > G_window_rows ()) ) {
		/* if cache size is invalid, just set to auto-mode (-1) */
		G_warning ("Invalid cache size requested (must be between 0 and %i or -1).\n", G_window_rows());
		cachesize = -1;
	}

		
	/* open logfile */
	if (strcmp ("none", parm.logfile->answer) != 0)
	{
		logfile = fopen (parm.logfile->answer, "w+");
		if (logfile == NULL)
		{
			G_fatal_error ("Logfile error: %s\n",
				       strerror (errno));
		}
		else
		{
			debug_mode = 1;
			fprintf (logfile,"This is r.dst.bpa, version %.2f\n",PROGVERSION);
			systime = time (NULL);
			fprintf (logfile,"Calculation started on %s",ctime(&systime));
			fprintf (logfile,"\tlocation   = %s\n",G_location());
			fprintf (logfile,"\tmapset     = %s\n",G_mapset());
			fprintf (logfile,"\tinput map  = %s\n",parm.input->answer);
			fprintf (logfile,"\tsites file = %s\n",parm.sites->answer);
			fprintf (logfile,"Results will be stored in '%s.<SITE|NOSITE|SITE_NOSITE>'\n\n",parm.output->answer);
			fflush (logfile);
		}
	}	
	
	/* do the actual calculations */
	write_BPAs (parm.input->answer, mapset, parm.sites->answer, parm.output->answer,
				atof(parm.input_uncertainty->answer), atof(parm.sites_uncertainty->answer),
				show_progress);
	
	/* write all that remains to be written to logfile */
	if (strcmp ("none", parm.logfile->answer) != 0) {
		/* write advisory if necessary */
		if ( WARN_ON ) {
			fprintf (logfile,"\nAdvice on this calculation:\n");
			if ( WARN_N_TO_SMALL ) {
				fprintf (logfile,"\tIncrease sample size to at least %i\n",MIN_SAMPLE_SIZE);
			}
		}
		/* write processing time to logfile */
		proctime = clock ();
		timeused = (unsigned long) proctime / CLOCKS_PER_SEC;
		days = timeused / 86400;
		hours = (timeused - (days * 86400)) / 3600;
		mins = (timeused - (days * 86400) - (hours * 3600)) / 60;		
		secs = (timeused - (days * 86400) - (hours * 3600) - (mins * 60));
		systime = time (NULL);
		fprintf (logfile,"\nCalculation finished on %s",ctime(&systime));		
		fprintf (logfile,"Processing time: %id, %ih, %im, %is\n",
				days, hours, mins, secs );
		fflush (logfile);
	}
	
	/* write some metadata to the output files' histories */
	G_short_history (parm.input->answer,"raster",&hist);
	
	strcpy (hist.datsrc_1, parm.input->answer);
	strcpy (hist.keywrd, "Generated by r.dst.bpa.");
	sprintf (result_str,"uncertainty(sites)=%s, uncertainty(area)=%s.",
				parm.sites_uncertainty->answer, parm.input_uncertainty->answer);	
	hist.edlinecnt = 2;
	strcpy (hist.edhist[0], "BPA values for hypothesis {H1}='site present'.");
	strcpy (hist.edhist[1], result_str);		
	sprintf (result_str,"%s.SITE",parm.output->answer);
	G_write_history (result_str, &hist);
	
	sprintf (result_str,"uncertainty(sites)=%s, uncertainty(area)=%s.",
				parm.sites_uncertainty->answer, parm.input_uncertainty->answer);
	strcpy (hist.edhist[0], "BPA values for hypothesis {H2}='no site present'.");	
	strcpy (hist.edhist[1], result_str);		
	sprintf (result_str,"%s.NOSITE",parm.output->answer);
	G_write_history (result_str, &hist);
	
	sprintf (result_str,"uncertainty(sites)=%s, uncertainty(area)=%s.",
				parm.sites_uncertainty->answer, parm.input_uncertainty->answer);
	strcpy (hist.edhist[0], "BPA values for hypothesis {H1,H2}='site present OR no site present'.");					
	strcpy (hist.edhist[1], result_str);		
	sprintf (result_str,"%s.SITE_NOSITE",parm.output->answer);
	G_write_history (result_str, &hist);
			
	return 0;
}
