/* s.random.sample */
/* Generate a random subset of vector points. */

/* TODO 
   - specifying -a automatically disregards the map as well.
     Is that good behaviour?
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#include "gis.h"
#include "site.h"
#include "../../GT/rowcache.h"

#define PROGVERSION 1.0

extern int errno;
int cachesize;

int debug_mode = 0;		/* 1 to enable writing debug output to logfile */

/* checks if user-supplied parameters are valid */
int
check_parms (char *map_answer, char *input_answer, char *output_answer, int all )
{

	int processing_mode = 0;

	if ( map_answer != NULL ) {
		if ((G_find_file ("fcell", map_answer, "") == NULL)
	    	&& (G_find_file ("dcell", map_answer, "") == NULL)
	    	&& (G_find_file ("cell", map_answer, "") == NULL))	
		{
			/* no coverage map found */
			G_fatal_error
				("No raster map exists with the given name.");
		}
		
		/* specifying all means to disregard the map! */
		if ( all == 1 ) {
			G_warning ("Map was specified, but the -a flag disregards map values!\n Your results might not be what you expect.");
		}
	}

	/* check if sample sites exist */
	if (G_find_file ("site_lists", input_answer, "") == NULL)
	{
		G_fatal_error
			("No sample/sites list exists with the given name!");
	}	

	/* check if a legal file name was given for output map
	   this just overwrites existing sites */ 
	if (G_legal_filename (output_answer) == -1)
	{
		G_fatal_error
			("Name chosen for result map is not a valid file name!");
	}
		
	/* return raster mode to caller */
	if ( map_answer != NULL ) {
		return (G_raster_map_type (map_answer,""));
	}
	else return (-1);
}

/* creates a new site file in the user's mapset that contains */
/* 'percentage' sites randomly picked from 'sitesfile' */
/* useful to create randomised samples for statistical tests */
/* if 'samplename' is not NULL, file will be kept instead of deleted. */
/* TODO: MEM LEAK: 'tmpfile' must be free'd by caller */
void do_split_sample ( char *input, char *output, double percentage, char *map,  
						int all, int processing_mode, int quiet) {
    CELL *cellbuf;
	DCELL *dcellbuf;
	GT_Row_cache_t *cache;
	int fd;
	int i,j,k,l;
	int no_sites;
	int sites_tried = 0;
	FILE *fp_sites;
	FILE *fp_outsites;
	int dims;
	int cat;
	int strs;
	int dbls;
	Site *site;
	struct Cell_head region;
	int error;
	unsigned int *taken; /* this is an array of 0/1 which signals, if
	                       a certain site has already been 'drawn' */
	long row_idx, col_idx;
	
	/* get current region */
	G_get_window (&region);
	
	/* attempt to create new site file for output sites */
	if ( (fp_outsites = G_sites_open_new (output)) == NULL ) {
		G_fatal_error ("Could not open file to store split sample.\n");
	}
	/* open sites file with samples */
	fp_sites =
		G_sites_open_old (input, G_find_sites (input, ""));
	if (fp_sites == NULL)
	{
		G_fatal_error
			("Could not open site list file with samples.");
	}
	error = G_site_describe (fp_sites, &dims, &cat, &strs, &dbls);
	if (error < 0)
	{
		G_fatal_error
			("Could not determine format of sample site list.");
	}
	/* guess site file format */
	site = G_site_new_struct (cat, dims, strs, dbls);
	if (site == NULL)
	{
		G_fatal_error
			("Could not reproduce format of sample site list.");
	}
	/* get total number of sites */
	i = 0;
	while (G_site_get (fp_sites, site) == 0)
	{
		if ( all == 1 ) {
			/* sites are not limited to current region */
			i++;
		} else {
			/* count only those sites within current region */
			if ( G_site_in_region (site,&region) ) {
				i++;
			}
		}
		
	}
	
	k = ( ((float) i/100)) * percentage; /* k now has the number of sites wanted */
	
	if ( quiet != 1 ) {
		fprintf (stderr,"Creating randomised sample of size n = %i.\n",k);
	}
	
	/* now, we need to acquire exactly 'k' random sites that fall in NON-NULL */
	/* coverage raster cells. */	
	taken = G_calloc (i, sizeof (unsigned int));
	for ( l = 0; l < k; l ++ ) {
		taken[l] = 0;
	}
	no_sites = i; /* store this for later use */
	
	/* does user want to filter sites through a raster map? */
	if ( map != NULL) {
		/* open raster map */
		fd = G_open_cell_old (map, G_find_cell (map, ""));
		if (fd < 0)
		{
			G_fatal_error ("Could not open raster map for reading!\n");
		}
		/* allocate cache and buffer, according to type of coverage */
		if ( processing_mode == CELL_TYPE) {
			/* INT coverage */
			cache = (GT_Row_cache_t *) G_malloc (sizeof (GT_Row_cache_t));
			/* TODO: check error value */
			error = GT_RC_open (cache, cachesize, fd, CELL_TYPE);
			cellbuf = G_allocate_raster_buf (CELL_TYPE);			
		}
		if ( (processing_mode == FCELL_TYPE) || (processing_mode == DCELL_TYPE) ) {
			/* FP coverage */
			cache = (GT_Row_cache_t *) G_malloc (sizeof (GT_Row_cache_t));
			/* TODO: check error value */
			error = GT_RC_open (cache, cachesize, fd, DCELL_TYPE);
			dcellbuf = G_allocate_raster_buf (DCELL_TYPE);	
		}
	}
	
	srand ( ((int) time (NULL)) + getpid()); /* set seed for random number generator from system time and process ID*/
	i = 0;
	
	/* MAIN LOOP */
	while ( i < k ) {
		//fprintf (stderr, "i = %i\n",i);
		/* get a random index, but one that was not taken already */
		l = 0;
		while ( l == 0 ) {
			j = rand () % ( no_sites - 1 + 1) + 1; /* j now has the random position to try */
			if ( taken[j-1] == 0 ) {
				l = 1; /* exit loop */
			}
		}
		taken [j-1] = 1; /* mark this index as 'taken' */
		sites_tried ++; /* keep track of this so we do not enter an infinite loop */
		if ( sites_tried > no_sites ) {
			/* could not create a large enough sample */
			G_fatal_error ("Could not find enough sites for split sampling.\nDecrease split sample size or increase number of sites.\n");
		}			
		/* rewind sites file and retreive the random site */
		fclose (fp_sites);
		fp_sites = G_sites_open_old (input, G_find_sites (input, ""));
		for ( l = 1; l <= j; l ++ ) {
			/* get next site */
			if ( G_site_get (fp_sites, site)  != 0 ) {
				G_fatal_error ("Error reading sites for creating random sample.\n");
			}
		}		
		/* now, check if coverage under site is NON-NULL and within region */
		/* convert site northing to row! */		
		row_idx =
			(long) G_northing_to_row (site->north,
				  &region);
				
		col_idx =
			(long) G_easting_to_col (site->east,
				 &region);
		/* do region check, first... */
		if ( (all != 1) && (G_site_in_region (site,&region)) ) {
			/* then read row from cache and check for NULL */
			/* if required */
			if ( map != NULL ) {
				if ( processing_mode == CELL_TYPE ) {
					cellbuf = GT_RC_get (cache, row_idx);			
					if (!G_is_c_null_value(&cellbuf[col_idx])) {
						i ++;
						G_site_put (fp_outsites, site);
						fflush (stdout);
					}
				}
				if ( (processing_mode == FCELL_TYPE) || (processing_mode == DCELL_TYPE) ) {
					dcellbuf = GT_RC_get (cache, row_idx);
					if (!G_is_d_null_value(&dcellbuf[col_idx])) {
						i ++;
						G_site_put (fp_outsites, site);
						fflush (stdout);
					}
				}
			} else {
				i ++;
				G_site_put (fp_outsites, site);
				fflush (stdout);
			}
		}
		/* disregard region setting and map, if -a flag is given */
		if ( all == 1 ) {
			i ++;
			G_site_put (fp_outsites, site);
			fflush (stdout);
		}
		
		if ( quiet != 1 ) {
			G_percent(i,k,1);
		}
	}
	/* END OF MAIN LOOP */		
	
	fclose (fp_sites);
	fclose (fp_outsites);
	/* return name of temp file for the split sample */
	
	if ( map != NULL ) {
		/* close cache, free buffers! */
		GT_RC_close (cache);
		if ( processing_mode == CELL_TYPE ) {
			G_free (cellbuf);
		}
		if ( (processing_mode == FCELL_TYPE) || (processing_mode == DCELL_TYPE) ) {
			G_free (dcellbuf);
		}
		G_free (cache);
		}
}


int
main (int argc, char *argv[])
{
	struct GModule *module;
	struct
	{
		struct Option *input; /* name of input site map */
		struct Option *output; /* name of output site map */
		struct Option *size; /* size of sample (%input) */
		struct Option *map; /* optional raster to check for NULL */
		struct Option *cachesize;
	}
	parm;
	struct
	{
		struct Flag *quiet;
		struct Flag *all; /* disregard region settings */	
	}
	flag;
	int processing_mode = 0;

	/* setup some basic GIS stuff */
	G_gisinit (argv[0]);
	module = G_define_module ();
	module->description = "Generate a random sample from a sitemap.";
	/* do not pause after a warning message was displayed */
	G_sleep_on_error (0);

	
	/* Module Parameters: */
	
	/* site map with known sites */
	parm.input = G_define_option ();
	parm.input->key = "input";
	parm.input->type = TYPE_STRING;
	parm.input->required = YES;
	parm.input->description = "Sites map to draw sample from.";
	
	/* name of output map */
	parm.output = G_define_option ();
	parm.output->key = "output";
	parm.output->type = TYPE_STRING;
	parm.output->required = YES;
	parm.output->description = "Sites map to store random sample in.";

	/* percentage of input sites to include for random sample */
	parm.size = G_define_option ();
	parm.size->key = "size";
	parm.size->type = TYPE_DOUBLE;
	parm.size->required = YES;
	parm.size->description = "Size of the sample to generate in percent of input sites.";
	parm.size->options = "0-100";

	/* map = an optional raster map. Samples will not be generated */
	/*       in cells where this map fprintf (stderr, "HI\n");is NULL */
	parm.map = G_define_option ();
	parm.map->key = "raster";
	parm.map->type = TYPE_STRING;
	parm.map->required = NO;
	parm.map->description = "Sites will not fall on NULL cells in this raster map.";
	
	/* number of lines to store in cache */
	parm.cachesize = G_define_option ();
	parm.cachesize->key = "cachesize";
	parm.cachesize->type = TYPE_INTEGER;
	parm.cachesize->answer = "-1";
	parm.cachesize->required = NO;
	parm.cachesize->description = "Number of raster rows to store in cache (-1 for auto).";
	
	flag.all = G_define_flag ();
	flag.all->key = 'a';
	flag.all->description = "Also draw sites from outside the current region's limits.";

	flag.quiet = G_define_flag ();
	flag.quiet->key = 'q';
	flag.quiet->description = "Quiet operation: no progress display.";

	G_disable_interactive ();
	
	/* parse command line */
	if (G_parser (argc, argv))
	{
		exit (-1);
	}

	/* check if given parameters are valid */
	
	processing_mode =
		check_parms (parm.map->answer, parm.input->answer, parm.output->answer, flag.all->answer);				
	
	if ( parm.map->answer != NULL ) {
		/* set cachesize */
		cachesize = atoi (parm.cachesize->answer);
		if ( (cachesize<-1) || (cachesize > G_window_rows ()) ) {
			/* if cache size is invalid, just set to auto-mode (-1) */
			G_warning ("Invalid cache size requested (must be between 0 and %i or -1).\n", G_window_rows());
			cachesize = -1;
		}
	}
		
    do_split_sample ( parm.input->answer, parm.output->answer, atof (parm.size->answer),
	                  parm.map->answer, flag.all->answer, processing_mode, flag.quiet->answer );
	
	return 0;
}
